# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs, Tuan Chien

import os
import re
import shutil
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Tuple

import click

from observatory.platform.observatory_config import (
    AirflowConnection,
    AirflowVariable,
    Api,
    BackendType,
    CloudSqlDatabase,
    CloudStorageBucket,
    ElasticSearch,
    Environment,
    GoogleCloud,
    Observatory,
    ObservatoryConfig,
    Terraform,
    TerraformConfig,
    VirtualMachine,
    WorkflowsProject,
    is_fernet_key,
    is_secret_key,
    Config,
)
from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.jinja2_utils import render_template
from observatory.platform.utils.proc_utils import stream_process

# Terminal formatting
BOLD = "\033[1m"
END = "\033[0m"


class DefaultWorkflowsProject:
    """Get Workflows Project configuration for when it was selected via the installer script (editable type)."""

    @staticmethod
    def academic_observatory_workflows():
        package_name = "academic-observatory-workflows"
        package = module_file_path("academic_observatory_workflows.dags", nav_back_steps=-3)
        package_type = "editable"
        dags_module = "academic_observatory_workflows.dags"

        return WorkflowsProject(
            package_name=package_name, package=package, package_type=package_type, dags_module=dags_module
        )

    @staticmethod
    def oaebu_workflows():
        package_name = "oaebu-workflows"
        package = module_file_path("oaebu_workflows.dags", nav_back_steps=-3)
        package_type = "editable"
        dags_module = "oaebu_workflows.dags"

        return WorkflowsProject(
            package_name=package_name, package=package, package_type=package_type, dags_module=dags_module
        )


class GenerateCommand:
    def generate_local_config(self, config_path: str, *, editable: bool, workflows: List[str], oapi: bool):
        """Command line user interface for generating an Observatory Config config.yaml.

        :param config_path: the path where the config file should be saved.
        :param editable: Whether the observatory platform is editable.
        :param workflows: List of installer script installed workflows.
        :return: None
        """

        file_type = "Observatory Config"
        click.echo(f"Generating {file_type}...")

        workflows = InteractiveConfigBuilder.get_installed_workflows(workflows)
        config = ObservatoryConfig(workflows_projects=workflows)

        if editable:
            InteractiveConfigBuilder.set_editable_observatory_platform(config.observatory)

            if oapi:
                InteractiveConfigBuilder.set_editable_observatory_api(config.observatory)

        config.save(path=config_path)

        click.echo(f'{file_type} saved to: "{config_path}"')

    def generate_terraform_config(self, config_path: str, *, editable: bool, workflows: List[str], oapi: bool):
        """Command line user interface for generating a Terraform Config config-terraform.yaml.

        :param config_path: the path where the config file should be saved.
        :param editable: Whether the observatory platform is editable.
        :param workflows: List of installer script installed workflows.
        :return: None
        """

        file_type = "Terraform Config"
        click.echo(f"Generating {file_type}...")
        workflows = InteractiveConfigBuilder.get_installed_workflows(workflows)
        config = TerraformConfig(workflows_projects=workflows)

        if editable:
            InteractiveConfigBuilder.set_editable_observatory_platform(config.observatory)

            if oapi:
                InteractiveConfigBuilder.set_editable_observatory_api(config.observatory)

        config.save(path=config_path)

        click.echo(f'{file_type} saved to: "{config_path}"')

    def generate_workflows_project(self, project_path: str, package_name: str, author_name: str):
        """Create all directories, init files and a setup.cfg + setup.py file for a new workflows project.

        :param project_path: The path to the new project directory
        :param package_name: The name of the new project package
        :param author_name: The name of the author, used for readthedocs.
        :return: None.
        """
        # Get paths to folders
        dag_dst_dir = os.path.join(project_path, package_name, "dags")
        workflow_dst_dir = os.path.join(project_path, package_name, "workflows")
        workflow_test_dst_dir = os.path.join(workflow_dst_dir, "tests")
        schema_dst_dir = os.path.join(project_path, package_name, "database", "schema")
        doc_dst_dir = os.path.join(project_path, "docs", "workflows")

        # Make folders
        for path in [dag_dst_dir, workflow_dst_dir, workflow_test_dst_dir, doc_dst_dir, schema_dst_dir]:
            os.makedirs(path, exist_ok=True)

        # Make init files
        package_folder = os.path.join(project_path, package_name)
        database_folder = os.path.join(project_path, package_name, "database")
        init_paths = [
            package_folder,
            dag_dst_dir,
            workflow_dst_dir,
            workflow_test_dst_dir,
            database_folder,
            schema_dst_dir,
        ]
        for path in init_paths:
            if not os.path.isfile(path):
                open(os.path.join(path, "__init__.py"), "a").close()

        templates_dir = module_file_path("observatory.platform.cli.templates.generate_project")

        # Create setup.cfg file
        setup_cfg_template = os.path.join(templates_dir, "setup.cfg.jinja2")
        setup_cfg = render_template(setup_cfg_template, package_name=package_name, python_version="3.7")
        setup_cfg_path = os.path.join(project_path, "setup.cfg")
        write_rendered_template(setup_cfg_path, setup_cfg, "setup.cfg")

        # Create setup.py file
        setup_py_template = os.path.join(templates_dir, "setup.py.jinja2")
        setup_py = render_template(setup_py_template, python_version="3.7")
        setup_py_path = os.path.join(project_path, "setup.py")
        write_rendered_template(setup_py_path, setup_py, "setup.py")

        # Create config.py with schema_folder function
        config_template = os.path.join(templates_dir, "config.py.jinja2")
        config = render_template(config_template, package_name=package_name)
        config_path = os.path.join(project_path, package_name, "config.py")
        write_rendered_template(config_path, config, "config.py")

        # Create a working docs directory using sphinx-quickstart
        create_docs_directory(project_path, package_name, author_name, templates_dir)

        print(
            f"""
        Created the following files and directories:
        
        └── {project_path}
            ├── docs
            │   ├── _build
            │   ├── _static
            │   ├── _templates
            │   ├── workflows
            │   ├── conf.py
            │   ├── generate_schema_csv.py
            │   ├── index.rst
            │   ├── make.bat
            │   ├── Makefile
            │   └── requirements.txt
            ├── {package_name}
            │   ├── dags
            │   │   └── __init__.py
            │   ├── database
            │   │   ├── schema
            │   │   │   └── __init__.py
            │   │   └── __init__.py
            │   └── workflows
            │   │   ├── tests
            │   │   │   └── __init__.py
            │   │   └── __init__.py
            │   ├── __init__.py
            │   └── config.py
            ├── setup.cfg
            └── setup.py
        """
        )

    def generate_workflow(self, project_path: str, package_name: str, workflow_type: str, workflow_class: str):
        """
        Write files for a new telescope which is using one of the templates

        :param project_path: the path to the workflows project.
        :param package_name: the Python package name.
        :param workflow_type: Type of telescope to generate.
        :param workflow_class: Class name of the new telescope.
        """
        # Add underscores between capitalised letters, make them all lowercase and strip underscores at the start/end
        workflow_module = re.sub(r"([A-Z])", r"_\1", workflow_class).lower().strip("_")

        # Get paths to folders
        dag_dst_dir = os.path.join(project_path, package_name, "dags")
        workflow_dst_dir = os.path.join(project_path, package_name, "workflows")
        schema_dst_dir = os.path.join(project_path, package_name, "database", "schema")
        doc_dst_dir = os.path.join(project_path, "docs")

        # Get paths to files
        dag_dst_file = os.path.join(dag_dst_dir, f"{workflow_module}.py")
        identifiers_dst_file = os.path.join(project_path, package_name, "identifiers.py")
        workflow_dst_file = os.path.join(workflow_dst_dir, f"{workflow_module}.py")
        test_dst_file = os.path.join(workflow_dst_dir, "tests", f"test_{workflow_module}.py")
        index_dst_file = os.path.join(doc_dst_dir, "index.rst")
        doc_dst_file = os.path.join(doc_dst_dir, "workflows", f"{workflow_module}.md")
        schema_dst_file = os.path.join(schema_dst_dir, f"{workflow_module}_{datetime.now().strftime('%Y-%m-%d')}.json")

        # Render templates
        workflow_path, dag_path, test_path, doc_index_path, doc_path, schema_path = get_workflow_template_path(
            workflow_type
        )
        dag = render_template(
            dag_path, workflow_module=workflow_module, workflow_class=workflow_class, package_name=package_name
        )
        workflow = render_template(
            workflow_path, workflow_module=workflow_module, workflow_class=workflow_class, package_name=package_name
        )
        test = render_template(
            test_path, workflow_module=workflow_module, workflow_class=workflow_class, package_name=package_name
        )
        doc = render_template(doc_path, workflow_module=workflow_module, workflow_class=workflow_class)
        schema = render_template(schema_path)

        # Write out files
        write_rendered_template(dag_dst_file, dag, "dag")
        write_rendered_template(workflow_dst_file, workflow, "workflow")
        write_rendered_template(test_dst_file, test, "test")
        write_rendered_template(doc_dst_file, doc, "documentation")
        write_rendered_template(schema_dst_file, schema, "schema")

        # Update documentation index
        with open(index_dst_file, "a") as f:
            f.write(f"   workflows/{workflow_module}\n")
        print(f"- Updated the documentation index file: {index_dst_file}")

        # Update TelescopeTypes in identifiers.py when using organisation template
        if workflow_type == "OrganisationTelescope":
            if not os.path.isfile(identifiers_dst_file):
                with open(identifiers_dst_file, "w") as f:
                    f.write("class WorkflowTypes:\n")
            with open(identifiers_dst_file, "a") as f:
                f.write(f'    {workflow_module} = "{workflow_module}"\n')
            print(f"- Updated the identifiers file: {identifiers_dst_file}")

    def generate_local_config_interactive(self, config_path: str, *, workflows: List[str], oapi: bool, editable: bool):
        """Construct an Observatory local config file through user assisted configuration.

        :param config_path: Configuration file path.
        :param workflows: List of installer script installed workflows projects.
        :param oapi: Whether installer script installed the Observatory API.
        :param editable: Whether the observatory platform is editable.
        """

        file_type = "Observatory Config"
        click.echo(f"Generating {file_type}...")

        config = InteractiveConfigBuilder.build(
            backend_type=BackendType.local, workflows=workflows, oapi=oapi, editable=editable
        )

        if editable:
            InteractiveConfigBuilder.set_editable_observatory_platform(config.observatory)

        config.save(config_path)
        click.echo(f'{file_type} saved to: "{config_path}"')

    def generate_terraform_config_interactive(
        self, config_path: str, *, workflows: List[str], oapi: bool, editable: bool
    ):
        """Construct an Observatory Terraform config file through user assisted configuration.

        :param config_path: Configuration file path.
        :param workflows: List of workflows projects installed by installer script.
        :param oapi: Whether installer script installed the Observatory API.
        :param editable: Whether the observatory platform is editable.
        """

        file_type = "Terraform Config"
        click.echo(f"Generating {file_type}...")

        config = InteractiveConfigBuilder.build(
            backend_type=BackendType.terraform, workflows=workflows, oapi=oapi, editable=editable
        )

        if editable:
            InteractiveConfigBuilder.set_editable_observatory_platform(config.observatory)

        config.save(config_path)
        click.echo(f'{file_type} saved to: "{config_path}"')


def get_workflow_template_path(workflow_type: str) -> Tuple[str, str, str, str, str, str]:
    """
    Get the correct template files to use.

    :param workflow_type: Name of the workflow type.
    :return: The template paths for the workflow, dag, test, doc and schema files
    """

    templates_dir = module_file_path("observatory.platform.workflows.templates")

    workflow_types = {
        "Workflow": {"dag": "dag.py.jinja2", "workflow": "workflow.py.jinja2", "test": "test.py.jinja2"},
        "SnapshotTelescope": {
            "dag": "dag.py.jinja2",
            "workflow": "workflow_snapshot.py.jinja2",
            "test": "test_snapshot.py.jinja2",
        },
        "StreamTelescope": {
            "dag": "dag_stream.py.jinja2",
            "workflow": "workflow_stream.py.jinja2",
            "test": "test_stream.py.jinja2",
        },
        "OrganisationTelescope": {
            "dag": "dag_organisation.py.jinja2",
            "workflow": "workflow_organisation.py.jinja2",
            "test": "test_organisation.py.jinja2",
        },
    }

    workflow_files = workflow_types.get(workflow_type)
    if workflow_files is None:
        raise Exception(f"Unsupported workflow type: {workflow_type}")

    workflow_file = workflow_files["workflow"]
    dag_file = workflow_files["dag"]
    test_file = workflow_files["test"]
    doc_index_file = "index.rst.jinja2"
    doc_file = "doc.md.jinja2"
    schema_file = "schema.json.jinja2"

    workflow_path = os.path.join(templates_dir, workflow_file)
    dag_path = os.path.join(templates_dir, dag_file)
    test_path = os.path.join(templates_dir, test_file)
    doc_index_path = os.path.join(templates_dir, doc_index_file)
    doc_path = os.path.join(templates_dir, doc_file)
    schema_path = os.path.join(templates_dir, schema_file)
    return workflow_path, dag_path, test_path, doc_index_path, doc_path, schema_path


def write_rendered_template(file_path: str, template: str, file_type: str):
    """Write the rendered template for a workflow file to a local file.

    :param file_path: The path to the local file.
    :param template: The rendered template.
    :param file_type: The file type, used for printing information.
    :return: None.
    """
    if os.path.exists(file_path):
        if not click.confirm(
            f"\nA {file_type} file already exists at: '{file_path}'\n" f"Would you like to overwrite the file?"
        ):
            return
    with open(file_path, "w") as f:
        f.write(template)
    print(f"- Created a new {file_type} file: {file_path}")


def create_docs_directory(project_path: str, package_name: str, author_name: str, templates_dir: str):
    """Create a new docs directory, part of a new workflows project

    :param project_path: The full path to the project source
    :param package_name: The name of the package inside the project
    :param author_name: The author name (required for sphinx quickstart)
    :param templates_dir: The directory with template files
    :return: None.
    """
    # Copy generate_schema_csv.py to docs
    src = os.path.join(templates_dir, "generate_schema_csv.py")
    dst = os.path.join(project_path, "docs", "generate_schema_csv.py")
    shutil.copy(src, dst)

    # Copy test_generate_schema_csv.py to docs
    src = os.path.join(templates_dir, "test_generate_schema_csv.py")
    dst = os.path.join(project_path, "docs", "test_generate_schema_csv.py")
    shutil.copy(src, dst)

    # Copy requirements.txt to docs
    src = os.path.join(templates_dir, "docs_requirements.txt")
    dst = os.path.join(project_path, "docs", "requirements.txt")
    shutil.copy(src, dst)

    # Run sphinx quickstart to set up docs
    if os.path.isfile(os.path.join(project_path, "docs", "conf.py")):
        print(
            f"WARNING, the docs directory already has configuration files. The sphinx-quickstart command to set "
            f"up the docs directory will raise an error."
        )
    sphinx_template_dir = os.path.join(templates_dir, "sphinx-quickstart")
    proc = subprocess.Popen(
        [
            "sphinx-quickstart",
            "-q",
            "-t",
            sphinx_template_dir,
            "-p",
            package_name,
            "-a",
            author_name,
            "-d",
            f"package_name={package_name}",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=os.path.join(project_path, "docs"),
    )
    stream_process(proc, True)


class FernetKeyType(click.ParamType):
    """Fernet key type for click prompt.  Will validate the input against the is_fernet_key method."""

    name = "FernetKeyType"

    def convert(
        self, value: Any, param: Optional[click.core.Parameter] = None, ctx: Optional[click.core.Context] = None
    ) -> Any:
        valid, msg = is_fernet_key(value)
        if not valid:
            self.fail(f"Input is not a valid Fernet key. Reason: {msg}", param=param, ctx=ctx)

        return value


class FlaskSecretKeyType(click.ParamType):
    """Secret key type for click prompt.  Will validate the input against the is_secret_key method."""

    name = "SecretKeyType"

    def convert(
        self, value: Any, param: Optional[click.core.Parameter] = None, ctx: Optional[click.core.Context] = None
    ) -> Any:
        valid, msg = is_secret_key(value)
        if not valid:
            self.fail(f"Input is not a valid secret key. Reason: {msg}", param=param, ctx=ctx)

        return value


class InteractiveConfigBuilder:
    """Helper class for configuring the ObservatoryConfig class parameters through interactive user input."""

    @staticmethod
    def set_editable_observatory_platform(observatory: Observatory):
        """Set observatory package settings to editable.

        :param observatory: Observatory object to change.
        """

        observatory.package = module_file_path("observatory.platform", nav_back_steps=-3)
        observatory.package_type = "editable"

    @staticmethod
    def set_editable_observatory_api(observatory: Observatory):
        """Set observatory api package settings to editable.

        :param observatory: Observatory object to change.
        """

        observatory.api_package = module_file_path("observatory.api", nav_back_steps=-3)
        observatory.api_package_type = "editable"

    @staticmethod
    def get_installed_workflows(workflows: List[str]) -> List[WorkflowsProject]:
        """Add the workflows projects installed by the installer script.

        :param workflows: List of installed workflows (via installer script).
        :return: List of WorkflowsProjects installed by the installer.
        """

        workflows_projects = []
        if "academic-observatory-workflows" in workflows:
            workflows_projects.append(DefaultWorkflowsProject.academic_observatory_workflows())

        if "oaebu-workflows" in workflows:
            workflows_projects.append(DefaultWorkflowsProject.oaebu_workflows())

        return workflows_projects

    @staticmethod
    def build(*, backend_type: BackendType, workflows: List[str], oapi: bool, editable: bool) -> Config:
        """Build the correct observatory configuration object through user assisted parameters.

        :param backend_type: The type of Observatory backend being configured.
        :param workflows: List of workflows installed by installer script.
        :param oapi: Whether installer script installed the Observatory API.
        :param editable: Whether the observatory platform is editable.
        :return: An observatory configuration object.
        """

        workflows_projects = InteractiveConfigBuilder.get_installed_workflows(workflows)

        if backend_type == BackendType.local:
            config = ObservatoryConfig(workflows_projects=workflows_projects)
        else:
            config = TerraformConfig(workflows_projects=workflows_projects)

        # Common sections for all backends
        InteractiveConfigBuilder.config_backend(config=config, backend_type=backend_type)
        InteractiveConfigBuilder.config_observatory(config=config, oapi=oapi, editable=editable)
        InteractiveConfigBuilder.config_terraform(config)
        InteractiveConfigBuilder.config_google_cloud(config)
        InteractiveConfigBuilder.config_airflow_connections(config)
        InteractiveConfigBuilder.config_airflow_variables(config)
        InteractiveConfigBuilder.config_workflows_projects(config)

        # Extra sections for Terraform
        if backend_type == BackendType.terraform:
            InteractiveConfigBuilder.config_cloud_sql_database(config)
            InteractiveConfigBuilder.config_airflow_main_vm(config)
            InteractiveConfigBuilder.config_airflow_worker_vm(config)
            InteractiveConfigBuilder.config_elasticsearch(config)
            InteractiveConfigBuilder.config_api(config)

        return config

    @staticmethod
    def config_backend(*, config: Config, backend_type: BackendType):
        """Configure the backend section.

        :param config: Configuration object to edit.
        :param backend_type: The backend type being used.
        """

        click.echo("Configuring backend settings")
        config.backend.type = backend_type

        text = "What kind of environment is this?"
        default = Environment.develop.name
        choices = click.Choice(
            choices=[Environment.develop.name, Environment.staging.name, Environment.production.name],
            case_sensitive=False,
        )

        config.backend.environment = Environment[
            click.prompt(text=text, type=choices, default=default, show_default=True, show_choices=True)
        ]

    @staticmethod
    def config_observatory(*, config: Config, oapi: bool, editable: bool):
        """Configure the observatory section.

        :param config: Configuration object to edit.
        """

        click.echo("Configuring Observatory settings")

        if editable:
            InteractiveConfigBuilder.set_editable_observatory_platform(config.observatory)
        # else:
        #     # Fill in if used installer script
        #     text = "What type of observatory platform installation did you perform? A git clone is an editable type, and a pip install is a pypi type."
        #     choices = click.Choice(choices=["editable", "sdist", "pypi"], case_sensitive=False)
        #     default = "pypi"
        #     package_type = click.prompt(text=text, type=choices, default=default, show_default=True, show_choices=True)
        #     config.observatory.package_type = package_type

        text = "Enter an Airflow Fernet key (leave blank to autogenerate)"
        default = ""
        fernet_key = click.prompt(text=text, type=FernetKeyType(), default=default)

        if fernet_key != "":
            config.observatory.airflow_fernet_key = fernet_key

        text = "Enter an Airflow secret key (leave blank to autogenerate)"
        default = ""
        secret_key = click.prompt(text=text, type=FlaskSecretKeyType(), default=default)

        if secret_key != "":
            config.observatory.airflow_secret_key = secret_key

        text = "Enter an email address to use for logging into the Airflow web interface"
        default = config.observatory.airflow_ui_user_email
        user_email = click.prompt(text=text, type=str, default=default, show_default=True)
        config.observatory.airflow_ui_user_email = user_email

        text = f"Password for logging in with {user_email}"
        default = config.observatory.airflow_ui_user_password
        user_pass = click.prompt(text=text, type=str, default=default, show_default=True)
        config.observatory.airflow_ui_user_password = user_pass

        text = "Enter observatory config directory. If it does not exist, it will be created."
        default = config.observatory.observatory_home
        observatory_home = click.prompt(
            text=text, type=click.Path(exists=False, readable=True), default=default, show_default=True
        )
        config.observatory.observatory_home = observatory_home
        Path(observatory_home).mkdir(exist_ok=True, parents=True)

        text = "Enter postgres password"
        default = config.observatory.postgres_password
        postgres_password = click.prompt(text=text, type=str, default=default, show_default=True)
        config.observatory.postgres_password = postgres_password

        text = "Redis port"
        default = config.observatory.redis_port
        redis_port = click.prompt(text=text, type=int, default=default, show_default=True)
        config.observatory.redis_port = redis_port

        text = "Flower UI port"
        default = config.observatory.flower_ui_port
        flower_ui_port = click.prompt(text=text, type=int, default=default, show_default=True)
        config.observatory.flower_ui_port = flower_ui_port

        text = "Airflow UI port"
        default = config.observatory.airflow_ui_port
        airflow_ui_port = click.prompt(text=text, type=int, default=default, show_default=True)
        config.observatory.airflow_ui_port = airflow_ui_port

        text = "Elastic port"
        default = config.observatory.elastic_port
        elastic_port = click.prompt(text=text, type=int, default=default, show_default=True)
        config.observatory.elastic_port = elastic_port

        text = "Kibana port"
        default = config.observatory.kibana_port
        kibana_port = click.prompt(text=text, type=int, default=default, show_default=True)
        config.observatory.kibana_port = kibana_port

        text = "API port"
        default = config.observatory.api_port
        api_port = click.prompt(text=text, type=int, default=default, show_default=True)
        config.observatory.api_port = api_port

        text = "Docker network name"
        default = config.observatory.docker_network_name
        docker_network_name = click.prompt(text=text, type=str, default=default, show_default=True)
        config.observatory.docker_network_name = docker_network_name

        text = "Is the docker network external?"
        default = config.observatory.docker_network_is_external
        docker_network_is_external = click.prompt(text=text, type=bool, default=default, show_default=True)
        config.observatory.docker_network_is_external = docker_network_is_external

        text = "Docker compose project name"
        default = config.observatory.docker_compose_project_name
        docker_compose_project_name = click.prompt(text=text, type=str, default=default, show_default=True)
        config.observatory.docker_compose_project_name = docker_compose_project_name

        text = "Do you wish to enable ElasticSearch and Kibana?"
        choices = click.Choice(choices=["y", "n"], case_sensitive=False)
        default = "y"
        enable_elk = click.prompt(text=text, default=default, type=choices, show_default=True, show_choices=True)

        config.observatory.enable_elk = True if enable_elk == "y" else False

        # If installed by installer script, we can fill in details
        if oapi and editable:
            InteractiveConfigBuilder.set_editable_observatory_api(config.observatory)
        elif not oapi:
            text = "Observatory API package name"
            default = config.observatory.api_package
            api_package = click.prompt(text=text, type=str, default=default, show_default=True)
            config.observatory.api_package = api_package

            text = "Observatory API package type"
            default = config.observatory.api_package_type
            api_package_type = click.prompt(text=text, type=str, default=default, show_default=True)
            config.observatory.api_package_type = api_package_type

    @staticmethod
    def config_google_cloud(config: Config):
        """Configure the Google Cloud section.

        :param config: Configuration object to edit.
        """

        zone = None
        region = None
        buckets = list()

        if not config.schema["google_cloud"]["required"]:
            text = "Do you want to configure Google Cloud settings?"
            proceed = click.confirm(text=text, default=False, abort=False, show_default=True)
            if not proceed:
                return

        text = "Google Cloud Project ID"
        project_id = click.prompt(text=text, type=str)

        text = "Path to Google Service Account key file (json)"
        credentials = click.prompt(text=text, type=click.Path(exists=True, readable=True))

        text = "Data location"
        default = "us"
        data_location = click.prompt(text=text, type=str, default=default, show_default=True)

        if config.backend.type == BackendType.terraform:
            text = "Region"
            default = "us-west1"
            region = click.prompt(text=text, type=str, default=default, show_default=True)

            text = "Zone"
            default = "us-west1-a"
            zone = click.prompt(text=text, type=str, default=default, show_default=True)

        if config.backend.type == BackendType.local:
            text = "Download bucket name"
            download_bucket = click.prompt(text=text, type=str)

            text = "Transform bucket name"
            transform_bucket = click.prompt(text=text, type=str)

            buckets.append(CloudStorageBucket(id="download_bucket", name=download_bucket))
            buckets.append(CloudStorageBucket(id="transform_bucket", name=transform_bucket))

        config.google_cloud = GoogleCloud(
            project_id=project_id,
            credentials=credentials,
            region=region,
            zone=zone,
            data_location=data_location,
            buckets=buckets,
        )

    @staticmethod
    def config_terraform(config: Config):
        """Configure the Terraform section.

        :param config: Configuration object to edit.
        """

        if not config.schema["terraform"]["required"]:
            text = "Do you want to configure Terraform settings?"
            proceed = click.confirm(text=text, default=False, abort=False, show_default=True)
            if not proceed:
                return

        if config.backend.type == BackendType.local:
            suffix = " (leave blank to disable)"
            default = ""
        else:
            suffix = ""
            default = None

        text = f"Terraform organization name{suffix}"
        organization = click.prompt(text=text, type=str, default=default)

        if organization == "":
            return

        config.terraform = Terraform(organization=organization)

    @staticmethod
    def config_airflow_connections(config: Config):
        """Configure the Airflow connections section.

        :param config: Configuration object to edit.
        """

        click.echo("Configuring Airflow Connections")

        text = "Do you have any Airlfow connections you wish to add?"
        add_connections = click.confirm(text=text, default=False, abort=False, show_default=True)

        if not add_connections:
            return

        connections = list()
        while True:
            text = "Airflow connection name"
            conn_id = click.prompt(text=text, type=str)

            text = "Airflow connection value"
            conn_name = click.prompt(text=text, type=str)

            connections.append(AirflowConnection(name=conn_id, value=conn_name))

            text = "Do you wish to add another connection?"
            add_another = click.confirm(text=text, default=False, abort=False, show_default=True)

            if not add_another:
                break

        config.airflow_connections = connections

    @staticmethod
    def config_airflow_variables(config: Config):
        """Configure the Airflow variables section.

        :param config: Configuration object to edit.
        """

        click.echo("Configuring Airflow Variables")

        text = "Do you want to add Airflow variables?"
        add_variables = click.confirm(text=text, default=False, abort=False, show_default=True)

        if not add_variables:
            return

        variables = list()
        while True:
            text = "Name of Airflow variable"
            name = click.prompt(text=text, type=str)

            text = "Value of Airflow variable"
            value = click.prompt(text=text, type=str)

            variables.append(AirflowVariable(name=name, value=value))

            text = "Do you wish to add another variable?"
            add_another = click.confirm(text=text, default=False, abort=False, show_default=True)

            if not add_another:
                break

        config.airflow_variables = variables

    @staticmethod
    def config_workflows_projects(config: Config):
        """Configure the DAGs projects section.

        :param config: Configuration object to edit.
        """

        click.echo(
            "Configuring workflows projects. If you opted to install some workflows projects through the installer script then they will be automatically added to the config file for you. If not, e.g., if you installed via pip, you will need to add those projects manually now (or later)."
        )

        text = "Do you want to add workflows projects?"
        add_workflows_projects = click.confirm(text=text, default=False, abort=False, show_default=True)

        if not add_workflows_projects:
            return

        projects = list()
        while True:
            text = "Workflows package name"
            package_name = click.prompt(text=text, type=str)

            text = "Workflows package, either a local path to a Python source (editable), sdist, or PyPI package name and version"
            package = click.prompt(text=text, type=click.Path(exists=True, readable=True))

            text = "Package type"
            choices = click.Choice(choices=["editable", "sdist", "pypi"], case_sensitive=False)
            default = "editable"
            package_type = click.prompt(text=text, default=default, type=choices, show_default=True, show_choices=True)

            text = "Python import path to the module that contains the Apache Airflow DAGs to load"
            dags_module = click.prompt(text=text, type=str)

            projects.append(
                WorkflowsProject(
                    package_name=package_name,
                    package=package,
                    package_type=package_type,
                    dags_module=dags_module,
                )
            )

            text = "Do you wish to add another DAGs project?"
            add_another = click.confirm(text=text, default=False, abort=False, show_default=True)

            if not add_another:
                break

        config.workflows_projects.extend(projects)

    @staticmethod
    def config_cloud_sql_database(config: TerraformConfig):
        """Configure the cloud SQL database section.

        :param config: Configuration object to edit.
        """

        click.echo("Configuring the Google Cloud SQL Database")

        text = "Google CloudSQL db tier"
        default = "db-custom-2-7680"
        tier = click.prompt(text=text, type=str, default=default, show_default=True)

        text = "Google CloudSQL backup start time"
        default = "23:00"
        backup_start_time = click.prompt(text=text, type=str, default=default, show_default=True)

        config.cloud_sql_database = CloudSqlDatabase(
            tier=tier,
            backup_start_time=backup_start_time,
        )

    @staticmethod
    def config_airflow_main_vm(config: TerraformConfig):
        """Configure the Airflow main virtual machine section.

        :param config: Configuration object to edit.
        """

        click.echo(BOLD + "Configuring settings for the main VM that runs the Airflow scheduler and webserver" + END)

        text = "Machine type"
        default = "n2-standard-2"
        machine_type = click.prompt(text=text, type=str, default=default, show_default=True)

        text = "Disk size (GB)"
        default = 50
        disk_size = click.prompt(text=text, type=int, default=default, show_default=True)

        text = "Disk type"
        schema = config.schema["airflow_main_vm"]["schema"]
        default = "pd-ssd"
        choices = click.Choice(choices=schema["disk_type"]["allowed"], case_sensitive=False)
        disk_type = click.prompt(text=text, type=choices, show_choices=True, default=default, show_default=True)

        text = "Create VM? If yes, and you run Terraform apply, the vm will be created. Otherwise if false, and you run Terraform apply, the vm will be destroyed."
        create = click.confirm(text=text, default=True, abort=False, show_default=True)

        config.airflow_main_vm = VirtualMachine(
            machine_type=machine_type,
            disk_size=disk_size,
            disk_type=disk_type,
            create=create,
        )

    @staticmethod
    def config_airflow_worker_vm(config: TerraformConfig):
        """Configure the Airflow worker virtual machine section.

        :param config: Configuration object to edit.
        """

        click.echo(BOLD + "Configuring settings for the worker VM" + END)

        text = "Machine type"
        default = "n1-standard-8"
        machine_type = click.prompt(text=text, type=str, default=default, show_default=True)

        text = "Disk size (GB)"
        default = 3000
        disk_size = click.prompt(text=text, type=int, default=default, show_default=True)

        text = "Disk type"
        schema = config.schema["airflow_worker_vm"]["schema"]
        default = "pd-standard"
        choices = click.Choice(choices=schema["disk_type"]["allowed"], case_sensitive=False)
        disk_type = click.prompt(text=text, type=choices, show_choices=True, default=default, show_default=True)

        text = "Create VM? If yes, and you run Terraform apply, the vm will be created. Otherwise if false, and you run Terraform apply, the vm will be destroyed."
        create = click.confirm(text=text, default=False, abort=False, show_default=True)

        config.airflow_worker_vm = VirtualMachine(
            machine_type=machine_type,
            disk_size=disk_size,
            disk_type=disk_type,
            create=create,
        )

    @staticmethod
    def config_elasticsearch(config: TerraformConfig):
        """Configure the ElasticSearch section.

        :param config: Configuration object to edit.
        """

        click.echo("Configuring ElasticSearch")

        text = "Elasticsearch host url, e.g., https://host:port"
        host = click.prompt(text=text, type=str)

        text = "API key"
        api_key = click.prompt(text=text, type=str)

        config.elasticsearch = ElasticSearch(
            host=host,
            api_key=api_key,
        )

    @staticmethod
    def config_api(config: TerraformConfig):
        """Configure the API section.

        :param config: Configuration object to edit.
        """

        click.echo("Configuring the Observatory API")

        text = "Custom domain name for the API, used for the Google Cloud Endpoints service"
        default = "api.observatory.academy"
        domain_name = click.prompt(text=text, type=str, default=default, show_default=True)

        text = "Subdomain scheme"
        default = "project_id"
        choices = click.Choice(choices=["project_id", "environment"], case_sensitive=False)
        subdomain = click.prompt(text=text, type=choices, default=default, show_default=True, show_choices=True)

        text = "Observatory API Docker Image"
        default = "us-docker.pkg.dev/academic-observatory/observatory-platform/observatory-api:latest"
        api_image = click.prompt(text=text, type=str, default=default, show_default=True)

        text = "Endpoints Runtime Docker Image"
        default = "gcr.io/endpoints-release/endpoints-runtime-serverless:2"
        er_image = click.prompt(text=text, type=str, default=default, show_default=True)

        config.api = Api(
            domain_name=domain_name,
            subdomain=subdomain,
            api_image=api_image,
            er_image=er_image,
        )



# Copyright 2020-2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

from __future__ import annotations

import glob
import json
import logging
import os
import re
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from multiprocessing import cpu_count
from typing import Callable, Dict, List, Optional, Tuple

import elasticsearch.exceptions
import google.cloud.bigquery as bigquery
import pendulum
from airflow.exceptions import AirflowException
import airflow.hooks.base
from airflow.models import TaskInstance
from airflow.sensors.external_task import ExternalTaskSensor
from natsort import natsorted
from observatory.platform.elastic.elastic import (
    Elastic,
    KeepInfo,
    KeepOrder,
    make_sharded_index,
)
from observatory.platform.elastic.kibana import Kibana, ObjectType, TimeField
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.file_utils import (
    load_file,
    write_to_file,
    yield_csv,
    yield_jsonl,
)
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    download_blobs_from_cloud_storage,
    select_table_shard_dates,
)
from observatory.platform.utils.workflow_utils import delete_old_xcoms
from observatory.platform.workflows.snapshot_telescope import SnapshotRelease
from observatory.platform.workflows.workflow import Workflow
from pendulum import Date


CSV_TYPES = ["csv", "csv.gz"]
JSONL_TYPES = ["jsonl", "jsonl.gz"]


def default_index_keep_info():
    """Construct a default lookup dictionary for index_keep_info settings."""

    keep_info = {"": KeepInfo(ordering=KeepOrder.newest, num=2)}
    return keep_info


@dataclass
class ElasticImportConfig:
    dag_id: str
    project_id: str
    dataset_id: str
    bucket_name: str
    elastic_conn_key: str
    kibana_conn_key: str
    data_location: str = None
    file_type: str = None
    sensor_dag_ids: List[str] = None
    elastic_mappings_path: str = None
    elastic_mappings_func: Callable = None
    kibana_spaces: List[str] = None
    kibana_time_fields: List[TimeField] = None
    index_keep_info: Dict[str, KeepInfo] = field(default_factory=default_index_keep_info)


def load_elastic_mappings_simple(path: str, table_prefix: str) -> Dict:
    """Load the Elastic mappings for a given table prefix.

    :param path: the path to the Elastic mappings.
    :param table_prefix: the table prefix.
    :return: the rendered mapping as a Dict.
    """

    return json.loads(load_file(os.path.join(path, f"{make_index_prefix(table_prefix)}-mappings.json")))


def make_index_prefix(table_id: str):
    """Convert a table_id into an Elastic / Kibana index.

    :param table_id: the table_id.
    :return: the Elastic / Kibana index.
    """

    return table_id.replace("_", "-")


def make_table_prefix(table_id: str):
    """Remove the date from a table_id.

    :param table_id: the table id including a shard date.
    :return: the table id.
    """

    # -8 is removing the date from the string.
    return table_id[:-8]


def extract_table_id(table_id: str) -> Tuple[str, Optional[str]]:
    """Extract the table_id and and the date from an index.

    :param table_id: the table id.
    :return: table_id and date.
    """

    results = re.search(r"\d{8}$", table_id)

    if results is None:
        return table_id, None

    return make_table_prefix(table_id), results.group(0)


def list_table_ids(project_id: str, dataset_id: str, release_date: pendulum.DateTime) -> List[str]:
    """List all of the table_ids within a BigQuery dataset.

    :param project_id: the Google Cloud project id.
    :param dataset_id: the BigQuery dataset id.
    :param release_date: the release date.
    :return: the table ids.
    """

    src_client = bigquery.Client(project=project_id)

    table_ids = []
    table_id_set = set()
    tables = src_client.list_tables(dataset_id, max_results=10000)
    for table in tables:
        table_id, shard_date = extract_table_id(table.table_id)
        table_id_set.add(table_id)

    # For all sharded tables, find the latest version and add it to table_ids
    for table_id in table_id_set:
        table_dates = select_table_shard_dates(project_id, dataset_id, table_id, release_date)
        if len(table_dates):
            table_ids.append(bigquery_sharded_table_id(table_id, table_dates[0]))
        else:
            print(f"Error: {table_id}")

    return table_ids


def export_bigquery_table(
    project_id: str, dataset_id: str, table_id: str, location: str, file_type: str, destination_uri: str
) -> bool:
    """Export a BigQuery table.

    :param project_id: the Google Cloud project ID.
    :param dataset_id: the BigQuery dataset ID of the Observatory Platform dataset.
    :param table_id: the name of the BigQuery table.
    :param location: the location of the BigQuery dataset.
    :param file_type: the type of file to save the exported data as; csv or jsonl.
    :param destination_uri: the Google Cloud storage bucket destination URI.
    :return: whether the dataset was exported successfully or not.
    """

    # Set destination format
    if file_type in CSV_TYPES:
        destination_format = bigquery.DestinationFormat.CSV
    elif file_type in JSONL_TYPES:
        destination_format = bigquery.DestinationFormat.NEWLINE_DELIMITED_JSON
    else:
        raise ValueError(f"export_bigquery_table: file type '{file_type}' is not supported")

    # Create and run extraction job
    client = bigquery.Client()
    source_table_id = f"{project_id}.{dataset_id}.{table_id}"
    extract_job_config = bigquery.ExtractJobConfig()

    # Set gz compression if file type ends in .gz
    if file_type.endswith(".gz"):
        extract_job_config.compression = bigquery.Compression.GZIP

    extract_job_config.destination_format = destination_format
    extract_job: bigquery.ExtractJob = client.extract_table(
        source_table_id, destination_uri, job_config=extract_job_config, location=location
    )
    extract_job.result()

    return extract_job.state == "DONE"


def load_elastic_index(
    *,
    data_path: str,
    table_id: str,
    release_date: Date,
    elastic_mappings_folder: str,
    elastic_mappings_func: Callable,
    file_type: str,
    elastic_host: str,
    elastic_api_key_id: str,
    elastic_api_key: str,
    chunk_size: int,
    num_threads: int,
) -> bool:
    """Load an observatory index into Elasticsearch.

    :param data_path: the path to the data.
    :param table_id: the id of the table that will be loaded into Elasticsearch.
    :param release_date: the release date.
    :param file_type: the file type of the data that will be loaded.
    :param elastic_mappings_folder: the mappings path.
    :param elastic_mappings_func: the mappings Callable.
    :param elastic_host: the full Elasticsearch host.
    :param elastic_api_key_id: the Elastic API key id.
    :param elastic_api_key: the Elastic API key.
    :param chunk_size: the size of the batches to load.
    :param num_threads: the number of threads to use for loading.
    :return: whether the data loading successfully or not.
    """

    results = []

    # Break table_id into various properties
    table_prefix = make_table_prefix(table_id)
    index_prefix = make_index_prefix(table_prefix)

    # Fetch all files that should be loaded into this index
    file_pattern = os.path.join(data_path, f"{table_id}_*.{file_type}")
    file_paths = natsorted(glob.glob(file_pattern))

    # Load mappings file
    mappings = elastic_mappings_func(elastic_mappings_folder, table_prefix)

    if len(file_paths) == 0:
        # If no files found then set result to False
        results.append(False)
    else:
        # Load function
        if file_type in CSV_TYPES:
            load_func = yield_csv
        elif file_type in JSONL_TYPES:
            load_func = yield_jsonl
        else:
            raise ValueError(f"load_index: file type '{file_type}' is not supported")

        client = Elastic(
            host=elastic_host,
            chunk_size=chunk_size,
            thread_count=num_threads,
            api_key_id=elastic_api_key_id,
            api_key=elastic_api_key,
        )

        # Delete existing index
        index_id_sharded = make_sharded_index(index_prefix, release_date)
        client.delete_index(index_id_sharded)

        # Load files into index
        for file_path in file_paths:
            logging.info(f"Loading file: {file_path}")
            try:
                result = client.index_documents(index_id_sharded, mappings, load_func(file_path))
            except Exception as e:
                logging.error(f"Loading file error: {file_path}, {e}")
                result = False
            results.append(result)

    return all(results)


class ElasticImportRelease(SnapshotRelease):
    MAX_PARALLEL_QUERIES = 100

    def __init__(
        self,
        *,
        dag_id: str,
        release_date: pendulum.DateTime,
        dataset_id: str,
        file_type: str,
        table_ids: List,
        project_id: str,
        bucket_name: str,
        data_location: str,
        elastic_host: str,
        elastic_api_key_id: str,
        elastic_api_key: str,
        elastic_mappings_folder: str,
        elastic_mappings_func: Callable,
        kibana_host: str,
        kibana_api_key_id: str,
        kibana_api_key: str,
        kibana_spaces: List,
        kibana_time_fields: List[TimeField],
        chunk_size: int = 10000,
        num_threads: int = 2,
        num_workers: int = cpu_count(),
    ):
        super().__init__(dag_id, release_date, "", "", "")
        self.dataset_id = dataset_id
        self.file_type = file_type
        self.table_ids = table_ids
        self.project_id = project_id
        self.bucket_name = bucket_name
        self.data_location = data_location
        self.elastic_host = elastic_host
        self.elastic_api_key_id = elastic_api_key_id
        self.elastic_api_key = elastic_api_key
        self.elastic_mappings_folder = elastic_mappings_folder
        self.elastic_mappings_func = elastic_mappings_func
        self.kibana_host = kibana_host
        self.kibana_api_key_id = kibana_api_key_id
        self.kibana_api_key = kibana_api_key
        self.kibana_spaces = kibana_spaces
        self.kibana_time_fields = kibana_time_fields
        self.chunk_size = chunk_size
        self.num_threads = num_threads
        self.num_workers = num_workers
        self.bucket_prefix = f"telescopes/{dag_id}/{self.release_id}"
        self.elastic_import_task_state_path = os.path.join(self.extract_folder, "elastic_import_task_state.json")

    def get_keep_info(self, *, index: str, index_keep_info: Dict) -> KeepInfo:
        """Get the best KeepInfo for a given index prefix in self.index_keep_info.
        If there is an exact index match, it will return the info for that.  Otherwise it will keep looking for higher
        level matches.  If it completely fails to match, it will return the default configuration.

        Assumes index names are in the format: {level0 name}-{level1 name}-...-{leveln name}.
        Index should not include the -YYYYMMD suffix.

        :param index: Index pattern to match.
        :param index_keep_info: Dictionary of settings on how to retain Elastic indices.
        :return: KeepInfo for an index.
        """

        index_name = index
        while True:
            if index_name == "" and index_name not in index_keep_info:
                break

            if index_name in index_keep_info:
                return index_keep_info[index_name]

            end_ptr = index_name.rfind("-")
            index_name = index_name[:end_ptr]

        # Return a default configuration. Keep newest 2.
        return KeepInfo(ordering=KeepOrder.newest, num=2)

    def export_bigquery_tables(self) -> bool:
        """Export the BigQuery tables to Google Cloud Storage.

        :return: whether the tables were exported successfully.
        """

        # Calculate the number of parallel queries. Since all of the real work is done on BigQuery run each export task
        # in a separate thread so that they can be done in parallel.
        num_queries = min(len(self.table_ids), self.MAX_PARALLEL_QUERIES)

        results = []
        with ThreadPoolExecutor(max_workers=num_queries) as executor:
            futures = list()
            futures_msgs = {}
            for table_id in self.table_ids:
                destination_uri = f"gs://{self.bucket_name}/{self.bucket_prefix}/{table_id}_*.{self.file_type}"
                msg = f"Exporting table_id={table_id} to: {destination_uri}"
                logging.info(msg)
                future = executor.submit(
                    export_bigquery_table,
                    self.project_id,
                    self.dataset_id,
                    table_id,
                    self.data_location,
                    self.file_type,
                    destination_uri,
                )
                futures.append(future)
                futures_msgs[future] = msg

            # Wait for completed tasks
            for future in as_completed(futures):
                success = future.result()
                msg = futures_msgs[future]
                results.append(success)
                if success:
                    logging.info(f"Export success: {msg}")
                else:
                    logging.error(f"Export failed: {msg}")

        return all(results)

    def download_exported_data(self) -> bool:
        """Download the exported data from Cloud Storage.

        :return: whether the data was downloaded successfully or not.
        """

        return download_blobs_from_cloud_storage(self.bucket_name, self.bucket_prefix, self.download_folder)

    def read_import_state(self) -> List[str]:
        """Loads which tables have been indexed.

        :return: the list of table ids.
        """

        if os.path.isfile(self.elastic_import_task_state_path):
            return json.loads(load_file(self.elastic_import_task_state_path))
        return []

    def write_import_state(self, indexed_table_ids: List[str]):
        """Saves which tables have been indexed.

        :param indexed_table_ids: the table ids.
        :return: None.
        """

        write_to_file(json.dumps(indexed_table_ids), self.elastic_import_task_state_path)

    def import_to_elastic(self) -> bool:
        """Import data into Elasticsearch.

        :return: whether the data imported successfully or not.
        """

        results = []
        indexed_table_ids = self.read_import_state()
        logging.info(f"The following tables have already been indexed: {indexed_table_ids}")

        with ProcessPoolExecutor(max_workers=self.num_workers) as executor:
            futures = list()
            futures_msgs = {}

            # Decide what tables need indexing
            to_index_table_ids = list(set(self.table_ids) - set(indexed_table_ids))
            logging.info(f"The following tables will be indexed: {to_index_table_ids}")

            # Load each table into an Elastic index
            for table_id in to_index_table_ids:
                logging.info(f"Starting indexing task: {table_id}")
                future = executor.submit(
                    load_elastic_index,
                    data_path=self.download_folder,
                    table_id=table_id,
                    release_date=self.release_date,
                    elastic_mappings_folder=self.elastic_mappings_folder,
                    elastic_mappings_func=self.elastic_mappings_func,
                    file_type=self.file_type,
                    elastic_host=self.elastic_host,
                    elastic_api_key_id=self.elastic_api_key_id,
                    elastic_api_key=self.elastic_api_key,
                    chunk_size=self.chunk_size,
                    num_threads=self.num_threads,
                )

                futures.append(future)
                futures_msgs[future] = table_id

            # Wait for completed tasks
            for future in as_completed(futures):
                success = future.result()
                results.append(success)
                if success:
                    # Update the state of table_ids that have been indexed
                    table_id = futures_msgs[future]
                    indexed_table_ids.append(table_id)
                    self.write_import_state(indexed_table_ids)
                    logging.info(f"Loading index success: {table_id}")
                else:
                    logging.error(f"Loading index failed: {table_id}")

        return all(results)

    def update_elastic_aliases(self) -> bool:
        """Update the elasticsearch aliases.

        :return: whether the aliases updated correctly or not,
        """

        client = Elastic(host=self.elastic_host, api_key_id=self.elastic_api_key_id, api_key=self.elastic_api_key)

        # Make aliases and indexes
        aliases = []
        indexes = []
        for table_id in self.table_ids:
            table_prefix = make_table_prefix(table_id)
            alias = make_index_prefix(table_prefix)
            index = make_sharded_index(alias, self.release_date)
            aliases.append(alias)
            indexes.append(index)

        # Create actions for deleting all indexes currently assigned to aliases
        actions = []
        for alias in aliases:
            current_alias_indexes = client.get_alias_indexes(alias)
            for alias_index in current_alias_indexes:
                actions.append({"remove": {"index": alias_index, "alias": alias}})

        # Create actions to add new index aliases
        for alias, index in zip(aliases, indexes):
            actions.append({"add": {"index": index, "alias": alias}})

        # Update all aliases at once
        success = False
        try:
            result = client.es.indices.update_aliases(actions=actions)
            success = result.get("acknowledged", False)
        except elasticsearch.exceptions.NotFoundError:
            pass

        return success

    def get_kibana_time_field(self, index_pattern_id):
        time_field_name = None

        for time_field in self.kibana_time_fields:
            if re.match(time_field.pattern, index_pattern_id):
                time_field_name = time_field.field_name
                break

        return time_field_name

    def create_kibana_index_patterns(self) -> bool:
        """Create the Kibana index patterns.

        :return: whether the index patterns were created successfully or not.
        """

        kibana = Kibana(host=self.kibana_host, api_key_id=self.kibana_api_key_id, api_key=self.kibana_api_key)

        results = []
        for table_id in self.table_ids:
            table_prefix = make_table_prefix(table_id)
            index_pattern_id = make_index_prefix(table_prefix)
            attributes = {"title": index_pattern_id, "timeFieldName": self.get_kibana_time_field(index_pattern_id)}

            # Create an index pattern for each space
            for space_id in self.kibana_spaces:
                result = kibana.create_object(
                    ObjectType.index_pattern, index_pattern_id, attributes=attributes, space_id=space_id, exists_ok=True
                )
                results.append(result)

        return all(results)

    def delete_stale_indices(self, index_keep_info: Dict):
        """Delete all the stale indices.

        :param index_keep_info: Dictionary of settings on how to retain Elastic indices.
        """

        client = Elastic(host=self.elastic_host, api_key_id=self.elastic_api_key_id, api_key=self.elastic_api_key)

        # If you want to do a global cleanup of all indices, we can just call delete_stale_indices on index="*" instead
        indexed_table_ids = self.read_import_state()

        for table_id in indexed_table_ids:
            table_prefix = make_table_prefix(table_id)
            index_prefix = make_index_prefix(table_prefix)
            keep_info = self.get_keep_info(index=index_prefix, index_keep_info=index_keep_info)
            client.delete_stale_indices(index=f"{index_prefix}-*", keep_info=keep_info)


class ElasticImportWorkflow(Workflow):
    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        bucket_name: str,
        elastic_conn_key: str,
        kibana_conn_key: str,
        data_location: str = "us",
        file_type: str = "csv.gz",
        sensor_dag_ids: List[str] = None,
        elastic_mappings_folder: str = None,
        elastic_mappings_func: Callable = None,
        kibana_spaces: List[str] = None,
        kibana_time_fields: List[TimeField] = None,
        dag_id: Optional[str] = "elastic_import",
        start_date: Optional[pendulum.DateTime] = pendulum.datetime(2020, 11, 1),
        schedule_interval: Optional[str] = "@weekly",
        catchup: Optional[bool] = False,
        airflow_vars: List = None,
        airflow_conns: List = None,
        index_keep_info: Dict = None,
    ):
        """Create the ElasticImportWorkflow.

        :param project_id: the project id to import data from.
        :param dataset_id: the dataset id to import data from.
        :param bucket_name: the bucket name where the exported BigQuery data will be saved.
        :param elastic_conn_key: the key of the Airflow connection with elasticsearch info
        :param kibana_conn_key: the key of the Airflow connection with kibana info
        :param data_location: the location of?
        :param file_type:  the file type to import, can be csv or jsonl.
        :param sensor_dag_ids: a list of the DAG ids to wait for with sensors.
        :param kibana_spaces: the kibana spaces to update after Elastic indexes.
        :param dag_id: the DAG id.
        :param start_date: the start date.
        :param schedule_interval: the schedule interval.
        :param catchup: whether to catchup.
        :param airflow_vars: the required Airflow Variables.
        :param airflow_conns: the required Airflow Connections.
        :param index_keep_info: Index retention policy info.
        """

        if airflow_vars is None:
            airflow_vars = [AirflowVars.DATA_PATH]

        if airflow_conns is None:
            airflow_conns = [elastic_conn_key, kibana_conn_key]

        self.index_keep_info = index_keep_info if index_keep_info is not None else default_index_keep_info()

        # Initialise Telesecope base class
        super().__init__(
            dag_id=dag_id,
            start_date=start_date,
            schedule_interval=schedule_interval,
            catchup=catchup,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
        )

        self.project_id = project_id
        self.dataset_id = dataset_id
        self.bucket_name = bucket_name
        self.elastic_conn_key = elastic_conn_key
        self.kibana_conn_key = kibana_conn_key
        self.data_location = data_location
        self.file_type = file_type
        self.elastic_mappings_folder = elastic_mappings_folder
        self.elastic_mappings_func = elastic_mappings_func
        self.kibana_time_fields = kibana_time_fields

        self.sensor_dag_ids = sensor_dag_ids
        if sensor_dag_ids is None:
            self.sensor_dag_ids = []

        self.kibana_spaces = kibana_spaces
        if kibana_spaces is None:
            self.kibana_spaces = []

        # Add sensors
        with self.parallel_tasks():
            for ext_dag_id in self.sensor_dag_ids:
                sensor = ExternalTaskSensor(
                    task_id=f"{ext_dag_id}_sensor", external_dag_id=ext_dag_id, mode="reschedule"
                )
                self.add_operator(sensor)

        # Setup tasks
        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.list_release_info)

        # Tasks
        self.add_task(self.export_bigquery_tables)
        self.add_task(self.download_exported_data)
        self.add_task(self.import_to_elastic)
        self.add_task(self.update_elastic_aliases)
        self.add_task(self.delete_stale_indices)
        self.add_task(self.create_kibana_index_patterns)
        self.add_task(self.cleanup)

    def list_release_info(self, **kwargs):
        """List the table ids that should be exported.

        :param kwargs: the context passed from the BranchPythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return:
        """

        # Get release date
        release_date = kwargs["next_execution_date"].subtract(microseconds=1).date()
        table_ids = list_table_ids(self.project_id, self.dataset_id, release_date)

        # Push table ids and release date
        ti: TaskInstance = kwargs["ti"]
        ti.xcom_push(Workflow.RELEASE_INFO, {"release_date": release_date.format("YYYYMMDD"), "table_ids": table_ids})

        return True

    def make_release(self, **kwargs) -> ElasticImportRelease:
        """Make a release instance. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: A release instance or list of release instances
        """

        ti: TaskInstance = kwargs["ti"]
        record = ti.xcom_pull(
            key=Workflow.RELEASE_INFO, task_ids=self.list_release_info.__name__, include_prior_dates=False
        )

        release_date = pendulum.parse(record["release_date"])
        table_ids = record["table_ids"]

        # Get Airflow connections
        elastic_conn = airflow.hooks.base.BaseHook.get_connection(self.elastic_conn_key)
        elastic_host = f"{elastic_conn.conn_type}://{elastic_conn.host}:{elastic_conn.port}"
        elastic_api_key_id = elastic_conn.login
        elastic_api_key = elastic_conn.password

        kibana_conn = airflow.hooks.base.BaseHook.get_connection(self.kibana_conn_key)
        kibana_host = f"{kibana_conn.conn_type}://{kibana_conn.host}:{kibana_conn.port}"
        kibana_api_key_id = kibana_conn.login
        kibana_api_key = kibana_conn.password

        return ElasticImportRelease(
            dag_id=self.dag_id,
            release_date=release_date,
            dataset_id=self.dataset_id,
            file_type=self.file_type,
            table_ids=table_ids,
            project_id=self.project_id,
            bucket_name=self.bucket_name,
            data_location=self.data_location,
            elastic_host=elastic_host,
            elastic_api_key_id=elastic_api_key_id,
            elastic_api_key=elastic_api_key,
            elastic_mappings_folder=self.elastic_mappings_folder,
            elastic_mappings_func=self.elastic_mappings_func,
            kibana_host=kibana_host,
            kibana_api_key_id=kibana_api_key_id,
            kibana_api_key=kibana_api_key,
            kibana_spaces=self.kibana_spaces,
            kibana_time_fields=self.kibana_time_fields,
        )

    def export_bigquery_tables(self, release: ElasticImportRelease, **kwargs):
        """Export tables from BigQuery.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        success = release.export_bigquery_tables()
        if not success:
            raise AirflowException("export_bigquery_tables task: failed to export tables")

    def download_exported_data(self, release: ElasticImportRelease, **kwargs):
        """Download the exported data.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        success = release.download_exported_data()
        if not success:
            raise AirflowException(
                "download_exported_data task: data failed to download from " "Google Cloud Storage successfully"
            )

    def import_to_elastic(self, release: ElasticImportRelease, **kwargs):
        """Import the data into Elasticsearch.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        success = release.import_to_elastic()
        if not success:
            raise AirflowException("import_to_elastic task: failed to load Elasticsearch indexes")

    def update_elastic_aliases(self, release: ElasticImportRelease, **kwargs):
        """Update Elasticsearch aliases.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        success = release.update_elastic_aliases()
        if not success:
            raise AirflowException("update_elastic_aliases failed")

    def delete_stale_indices(self, release: ElasticImportRelease, **kwargs):
        """Delete stale Elasticsearch indices.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        release.delete_stale_indices(self.index_keep_info)

    def create_kibana_index_patterns(self, release: ElasticImportRelease, **kwargs):
        """Create Kibana index patterns.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        success = release.create_kibana_index_patterns()
        if not success:
            raise AirflowException("create_kibana_index_patterns failed")

    def cleanup(self, release: ElasticImportRelease, **kwargs):
        """Cleanup local files. Deletes old xcoms.

        :param release: the ElasticRelease.
        :param kwargs: the context passed from the Airflow Operator.
        See https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed
        to this argument.
        :return: None.
        """

        release.cleanup()

        execution_date = kwargs["execution_date"]
        delete_old_xcoms(dag_id=self.dag_id, execution_date=execution_date)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs, Tuan Chien

import os
import subprocess
import sys
from typing import ClassVar

import click

from observatory.platform.cli.cli_utils import INDENT1, INDENT2, INDENT3, indent
from observatory.platform.cli.generate_command import GenerateCommand
from observatory.platform.cli.platform_command import PlatformCommand
from observatory.platform.cli.terraform_command import TerraformCommand
from observatory.platform.docker.platform_runner import DEBUG, HOST_UID
from observatory.platform.observatory_config import Config, TerraformConfig, ObservatoryConfig
from observatory.platform.observatory_config import (
    generate_fernet_key,
    generate_secret_key,
)
from observatory.platform.utils.config_utils import observatory_home
from observatory.platform.utils.config_utils import (
    terraform_credentials_path as default_terraform_credentials_path,
)
from observatory.platform.utils.proc_utils import stream_process

PLATFORM_NAME = "Observatory Platform"
TERRAFORM_NAME = "Observatory Terraform"

LOCAL_CONFIG_PATH = os.path.join(observatory_home(), "config.yaml")
TERRAFORM_CONFIG_PATH = os.path.join(observatory_home(), "config-terraform.yaml")


def load_config(cls: ClassVar, config_path: str) -> Config:
    """Load a config file.
    :param cls: the config file class.
    :param config_path: the path to the config file.
    :return: the config file or exit with an OS.EX_CONFIG error or FileExistsError.
    """

    print(indent("config.yaml:", INDENT1))
    config_exists = os.path.exists(config_path)

    if not config_exists:
        msg = indent(f"- file not found, generating a default file on path: {config_path}", INDENT2)
        generate_cmd = GenerateCommand()
        if cls == ObservatoryConfig:
            print(msg)
            generate_cmd.generate_local_config(config_path, editable=False, workflows=[], oapi=False)
        elif cls == TerraformConfig:
            print(msg)
            generate_cmd.generate_terraform_config(config_path, editable=False, workflows=[], oapi=False)
        else:
            print(indent(f"- file not found, exiting: {config_path}", INDENT2))
        exit(os.EX_CONFIG)
    else:
        cfg = cls.load(config_path)
        if cfg.is_valid:
            print(indent("- file valid", INDENT2))
        else:
            print(indent("- file invalid", INDENT2))
            for error in cfg.errors:
                print(indent(f"- {error.key}: {error.value}", INDENT3))
            exit(os.EX_CONFIG)

        return cfg


@click.group()
def cli():
    """The Observatory Platform command line tool.

    COMMAND: the commands to run include:\n
      - platform: start and stop the local Observatory Platform platform.\n
      - generate: generate a variety of outputs.\n
      - terraform: manage Terraform Cloud workspaces.\n
    """

    pass


@cli.command()
@click.argument("command", type=click.Choice(["start", "stop"]))
@click.option(
    "--config-path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default=LOCAL_CONFIG_PATH,
    help="The path to the config.yaml configuration file.",
    show_default=True,
)
@click.option(
    "--host-uid",
    type=click.INT,
    default=HOST_UID,
    help="The user id of the host system. Used to set the user id in the Docker containers.",
    show_default=True,
)
@click.option("--debug", is_flag=True, default=DEBUG, help="Print debugging information.")
def platform(command: str, config_path: str, host_uid: int, debug):
    """Run the local Observatory Platform platform.\n

    COMMAND: the command to give the platform:\n
      - start: start the platform.\n
      - stop: stop the platform.\n
    """

    min_line_chars = 80
    print(f"{PLATFORM_NAME}: checking dependencies...".ljust(min_line_chars), end="\r")
    cfg = load_config(ObservatoryConfig, config_path)

    # Make the platform command, which encapsulates functionality for running the observatory
    platform_cmd = PlatformCommand(cfg, host_uid=host_uid, debug=debug)

    # Check dependencies
    platform_check_dependencies(platform_cmd, min_line_chars=min_line_chars)

    # Start the appropriate process
    if command == "start":
        platform_start(platform_cmd)
    elif command == "stop":
        platform_stop(platform_cmd)

    exit(os.EX_OK)


def platform_check_dependencies(platform_cmd: PlatformCommand, min_line_chars: int = 80):
    """Check Platform dependencies.

    :param platform_cmd: the platform command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    if not platform_cmd.is_environment_valid:
        print(f"{PLATFORM_NAME}: dependencies missing".ljust(min_line_chars))
    else:
        print(f"{PLATFORM_NAME}: all dependencies found".ljust(min_line_chars))

    print(indent("Docker:", INDENT1))
    if platform_cmd.docker_exe_path is not None:
        print(indent(f"- path: {platform_cmd.docker_exe_path}", INDENT2))

        if platform_cmd.is_docker_running:
            print(indent(f"- running", INDENT2))
        else:
            print(indent("- not running, please start", INDENT2))
    else:
        print(indent("- not installed, please install https://docs.docker.com/get-docker/", INDENT2))

    print(indent("Docker Compose:", INDENT1))
    if platform_cmd.docker_compose_path is not None:
        print(indent(f"- path: {platform_cmd.docker_compose_path}", INDENT2))
    else:
        print(indent("- not installed, please install https://docs.docker.com/compose/install/", INDENT2))

    if not platform_cmd.is_environment_valid:
        exit(os.EX_CONFIG)

    print(indent("Host machine settings:", INDENT1))
    print(indent(f"- observatory home: {platform_cmd.config.observatory.observatory_home}", INDENT2))
    print(indent(f"- host-uid: {platform_cmd.host_uid}", INDENT2))


def platform_start(platform_cmd: PlatformCommand, min_line_chars: int = 80):
    """Check Platform dependencies.

    :param platform_cmd: the platform command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    print(f"{PLATFORM_NAME}: building...".ljust(min_line_chars), end="\r")
    response = platform_cmd.build()

    if response.return_code == 0:
        print(f"{PLATFORM_NAME}: built".ljust(min_line_chars))
    else:
        print(f"{PLATFORM_NAME}: build error".ljust(min_line_chars))
        print(response.output)
        exit(os.EX_CONFIG)

    # Start the built containers
    print(f"{PLATFORM_NAME}: starting...".ljust(min_line_chars), end="\r")
    response = platform_cmd.start()

    if response.return_code == 0:
        ui_started = platform_cmd.wait_for_airflow_ui(timeout=120)

        if ui_started:
            print(f"{PLATFORM_NAME}: started".ljust(min_line_chars))
            print(f"View the Apache Airflow UI at {platform_cmd.ui_url}")
        else:
            print(f"{PLATFORM_NAME}: error starting".ljust(min_line_chars))
            print(f"Could not find the Airflow UI at {platform_cmd.ui_url}")
    else:
        print("Error starting the Observatory Platform")
        print(response.output)
        exit(os.EX_CONFIG)


def platform_stop(platform_cmd: PlatformCommand, min_line_chars: int = 80):
    """Start the Observatory platform.

    :param platform_cmd: the platform command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    print(f"{PLATFORM_NAME}: stopping...".ljust(min_line_chars), end="\r")
    response = platform_cmd.stop()

    if platform_cmd.debug:
        print(response.output)

    if response.return_code == 0:
        print(f"{PLATFORM_NAME}: stopped".ljust(min_line_chars))
    else:
        print(f"{PLATFORM_NAME}: error stopping".ljust(min_line_chars))
        print(response.error)
        exit(os.EX_CONFIG)


@cli.group()
def generate():
    """The Observatory Platform generate command.

    COMMAND: the commands to run include:\n
      - secrets: generate secrets.\n
      - config: generate configuration files for the Observatory Platform.\n
      - project: generate a new project directory and required files.\n
      - workflow: generate all files for a new workflow.
    """

    pass


@generate.command()
@click.argument("project_path", type=click.Path(exists=False, file_okay=False, dir_okay=True))
@click.argument("package_name", type=str)
@click.argument("author_name", type=str)
def project(project_path: str, package_name: str, author_name: str):
    """Generate a new workflows project.

    \b
    PROJECT_PATH is the Python project path.
    PACKAGE_NAME is the Python package name.
    AUTHOR_NAME is the author name, used for readthedocs.

    \b
    For example, the command: observatory generate project /path/to/my-workflows-project my_workflows_project

    \b Will generate the following files and folders:

    \b
    └── my-workflows-project
        ├── docs
        │   ├── _build
        │   ├── _static
        │   ├── _templates
        │   ├── workflows
        │   ├── conf.py
        │   ├── generate_schema_csv.py
        │   ├── index.rst
        │   ├── make.bat
        │   ├── Makefile
        │   ├── requirements.txt
        │   └── test_generate_schema_csv.py
        ├── {package_name}
        │   ├── dags
        │   │   └── __init__.py
        │   ├── database
        │   │   ├── schema
        │   │   │   └── __init__.py
        │   │   └── __init__.py
        │   └── workflows
        │   │   ├── tests
        │   │   │   └── __init__.py
        │   │   └── __init__.py
        │   ├── __init__.py
        │   └── config.py
        ├── setup.cfg
        └── setup.py
    """
    cmd = GenerateCommand()
    cmd.generate_workflows_project(project_path, package_name, author_name)

    if click.confirm(
        f"Would you like to install the '{package_name}' package inside your new project? This is required for "
        f"the workflows to function."
    ):
        proc = subprocess.Popen(
            [sys.executable, "-m", "pip", "install", "-e", "."],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=project_path,
        )
        stream_process(proc, True)


@generate.command()
@click.argument(
    "workflow_type", type=click.Choice(["Workflow", "StreamTelescope", "SnapshotTelescope", "OrganisationTelescope"])
)
@click.argument("workflow_name", type=str)
@click.option(
    "-p",
    "--project-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default=os.getcwd(),
    help="The path to the project directory.",
    show_default=True,
)
def workflow(workflow_type: str, workflow_name: str, project_path: str):
    """Generate all files for a new workflow.

    \b
    WORKFLOW_TYPE is the type of workflow.
    WORKFLOW_NAME is the the name of your new workflow.
    PROJECT_PATH is the Python project path, default is current directory.

    \b
    For example, the command: observatory generate workflow /path/to/my-workflows-project my_workflows_project Workflow MyWorkflow

    \b Will generate the files in bold inside your workflows project:

    \b
    └── my-workflows-project
        ├── docs
        │   ├── _build
        │   ├── _static
        │   ├── _templates
        │   ├── workflows
        │   │   └── \033[1mmy_workflow.md\033[0m
        │   ├── generate_schema_csv.py
        │   ├── index.rst
        │   ├── make.bat
        │   └── Makefile
        ├── my_workflows_project
        │   ├── dags
        │   │   ├── __init__.py
        │   │   └── \033[1mmy_workflow.py\033[0m
        │   ├── database
        │   │   ├── schema
        │   │   │   ├── __init__.py
        │   │   │   └── \033[1mmy_workflow_YYYY_MM_DD.json\033[0m
        │   │   └── __init__.py
        │   └── workflows
        │   │   ├── tests
        │   │   │   ├── __init__.py
        │   │   │   └── \033[1mtest_my_workflow.py\033[0m
        │   │   ├── __init__.py
        │   │   └── \033[1mmy_workflow.py\033[0m
        │   ├── __init__.py
        │   ├── config.py
        │   └── \033[1midentifiers.py\033[0m (OrganisationTelescope only)
        ├── setup.cfg
        └── setup.py
    """
    print(f"Given workflow type: {workflow_type}")
    print(f"Given workflow name: {workflow_name}")
    print(f"Given project path: {project_path}\n")

    # Check if egg-info dir is available
    egg_info_dir = [d for d in os.listdir(project_path) if d.endswith(".egg-info")]
    if not egg_info_dir:
        print(
            "Invalid workflows project, the given projects directory does not contain an installed python package.\n"
            "Either run this command from inside a valid workflows project or specify the path to a valid workflows "
            "project with the '--project-path' option.\n"
            "A new workflows project can be created using the 'observatory generate project' command."
        )
        exit(os.EX_CONFIG)
    if len(egg_info_dir) > 1:
        print(
            "Invalid workflows project, the given projects directory contains more than 1 installed python package.\n"
            "Either run this command from inside a valid workflows project or specify the path to a valid workflows "
            "project with the '--project-path' option.\n"
            "A new workflows project can be created using the 'observatory generate project' command."
        )
        exit(os.EX_CONFIG)

    # Get package name
    top_level_file = os.path.join(project_path, egg_info_dir[0], "top_level.txt")
    with open(top_level_file, "r") as f:
        package_name = f.readline().strip()
        print(f"Found package inside project: {package_name}")

    cmd = GenerateCommand()
    cmd.generate_workflow(
        project_path=project_path, package_name=package_name, workflow_type=workflow_type, workflow_class=workflow_name
    )


@generate.command()
@click.argument("command", type=click.Choice(["fernet-key", "secret-key"]))
def secrets(command: str):
    """Generate secrets for the Observatory Platform.\n

    COMMAND: the type of secret to generate:\n
      - fernet-key: generate a random Fernet Key.\n
      - secret-key: generate a random secret key.\n
    """

    if command == "fernet-key":
        print(generate_fernet_key())
    else:
        print(generate_secret_key())


@generate.command()
@click.argument("command", type=click.Choice(["local", "terraform"]))
@click.option(
    "--config-path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default=None,
    help="The path to the config file to generate.",
    show_default=True,
)
@click.option("--interactive", flag_value=True, help="Configuration through an interactive Q&A mode")
@click.option(
    "--ao-wf",
    flag_value=True,
    help="Indicates that the academic-observatory-workflows was installed through the cli installer script",
)
@click.option(
    "--oaebu-wf",
    flag_value=True,
    help="Indicates that the oaebu-workflows was installed through the cli installer script",
)
@click.option(
    "--oapi", flag_value=True, help="Indicates that the observatory api was installed through the cli installer script"
)
@click.option("--editable", flag_value=True, help="Indicates the observatory platform is editable")
def config(command: str, config_path: str, interactive: bool, ao_wf: bool, oaebu_wf: bool, oapi: bool, editable: bool):
    """Generate config files for the Observatory Platform.\n

    COMMAND: the type of config file to generate:\n
      - local: generate a config file for running the Observatory Platform locally.\n
      - terraform: generate a config file for running the Observatory Platform with Terraform.\n

    :param interactive: whether to interactively ask for configuration options.
    :param ao_wf: Whether academic_observatory_workflows was installed using the installer script.
    :param oaebu_wf: Whether oaebu_workflows was installed using the installer script.
    :param oapi: Whether the Observatory API was installed using the installer script.
    :param editable: Whether the observatory platform is editable.
    """

    # Make the generate command, which encapsulates functionality for generating data
    cmd = GenerateCommand()

    if config_path is None:
        config_path = LOCAL_CONFIG_PATH if command == "local" else TERRAFORM_CONFIG_PATH

    config_name = "Observatory config" if command == "local" else "Terraform config"

    workflows = []
    if ao_wf:
        workflows.append("academic-observatory-workflows")
    if oaebu_wf:
        workflows.append("oaebu-workflows")

    if command == "local" and not interactive:
        cmd_func = cmd.generate_local_config
    elif command == "terraform" and not interactive:
        cmd_func = cmd.generate_terraform_config
    elif command == "local" and interactive:
        cmd_func = cmd.generate_local_config_interactive
    else:
        cmd_func = cmd.generate_terraform_config_interactive

    if not os.path.exists(config_path) or click.confirm(
        f'The file "{config_path}" exists, do you want to overwrite it?'
    ):
        cmd_func(config_path, workflows=workflows, oapi=oapi, editable=editable)
    else:
        click.echo(f"Not generating {config_name}")


# increase content width for cleaner help output
@cli.command(context_settings=dict(max_content_width=120))
@click.argument(
    "command",
    type=click.Choice(["build-terraform", "build-image", "build-api-image", "create-workspace", "update-workspace"]),
)
# The path to the config-terraform.yaml configuration file.
@click.argument("config-path", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option(
    "--terraform-credentials-path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default=default_terraform_credentials_path(),
    help="",
    show_default=True,
)
@click.option("--debug", is_flag=True, default=DEBUG, help="Print debugging information.")
def terraform(command, config_path, terraform_credentials_path, debug):
    """Commands to manage the deployment of the Observatory Platform with Terraform Cloud.\n

    COMMAND: the type of config file to generate:\n
      - create-workspace: create a Terraform Cloud workspace.\n
      - update-workspace: update a Terraform Cloud workspace.\n
      - build-image: build a Google Compute image for the Terraform deployment with Packer.\n
      - build-terraform: build the Terraform files.\n
    """

    cfg = load_config(TerraformConfig, config_path)
    terraform_cmd = TerraformCommand(cfg, terraform_credentials_path, debug=debug)
    generate_cmd = GenerateCommand()

    # Check dependencies
    terraform_check_dependencies(terraform_cmd, generate_cmd)

    # Run commands
    if command == "build-terraform":
        # Build image with packer
        terraform_cmd.build_terraform()
    elif command == "build-image":
        # Build image with packer
        terraform_cmd.build_image()
    elif command == "build-api-image":
        # Build docker image stored in google container registry
        terraform_cmd.build_google_container_image()
    else:
        # Create a new workspace
        if command == "create-workspace":
            terraform_cmd.create_workspace()

        # Update an existing workspace
        elif command == "update-workspace":
            terraform_cmd.update_workspace()


def terraform_check_dependencies(
    terraform_cmd: TerraformCommand, generate_cmd: GenerateCommand, min_line_chars: int = 80
):
    """Check Terraform dependencies.

    :param terraform_cmd: the Terraform command instance.
    :param generate_cmd: the generate command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    print(f"{TERRAFORM_NAME}: checking dependencies...".ljust(min_line_chars), end="\r")

    if not terraform_cmd.is_environment_valid:
        print(f"{TERRAFORM_NAME}: dependencies missing".ljust(min_line_chars))
    else:
        print(f"{TERRAFORM_NAME}: all dependencies found".ljust(min_line_chars))

    print(indent("Terraform credentials file:", INDENT1))
    if terraform_cmd.terraform_credentials_exists:
        print(indent(f"- path: {terraform_cmd.terraform_credentials_path}", INDENT2))
    else:
        print(indent("- file not found, create one by running 'terraform login'", INDENT2))

    print(indent("Packer", INDENT1))
    if terraform_cmd.terraform_builder.packer_exe_path is not None:
        print(indent(f"- path: {terraform_cmd.terraform_builder.packer_exe_path}", INDENT2))
    else:
        print(indent("- not installed, please install https://www.packer.io/docs/install", INDENT2))

    print(indent("Google Cloud SDK", INDENT1))
    if terraform_cmd.terraform_builder.gcloud_exe_path is not None:
        print(indent(f"- path: {terraform_cmd.terraform_builder.gcloud_exe_path}", INDENT2))
    else:
        print(indent("- not installed, please install https://cloud.google.com/sdk/docs/install", INDENT2))

    if not terraform_cmd.is_environment_valid:
        exit(os.EX_CONFIG)


if __name__ == "__main__":
    cli()



# Copyright 2020, 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Tuan Chien


import logging
from dataclasses import dataclass
from enum import Enum
from multiprocessing import cpu_count
from typing import Dict, Iterable, Iterator, List

import elasticsearch.exceptions
import pendulum
from elasticsearch import Elasticsearch
from elasticsearch.helpers import parallel_bulk, scan
from pendulum import Date


def make_sharded_index(index_prefix: str, release_date: Date) -> str:
    """Make a sharded Elasticsearch index given an index prefix and a date.

    :param index_prefix: the index prefix.
    :param release_date: the date.
    :return: the sharded index.
    """

    return f'{index_prefix}-{release_date.strftime("%Y%m%d")}'


def make_elastic_uri(schema: str, user: str, secret: str, hostname: str, port: int) -> str:
    """Make an Elasticsearch URI.

    :param schema: the schema, e.g. http or https.
    :param user: Elasticsearch username.
    :param secret: Elasticsearch secret.
    :param hostname: Elasticsearch hostname.
    :param port: Elasticsearch port.
    :return: the full Elasticsearch URI.
    """

    return f"{schema}://{user}:{secret}@{hostname}:{port}"


class KeepOrder(Enum):
    """Specify to stale index cleaner whether to keep newest, or oldest indices."""

    newest = "newest"
    oldest = "oldest"


@dataclass
class KeepInfo:
    """Information on how do delete stale Elastic indices."""

    num: int = 2
    ordering: KeepOrder = KeepOrder.newest


class StaleIndexCleaner:
    """Helper class to delete stale elastic search indices."""

    @staticmethod
    def has_yyyymmdd_suffix(index: str) -> bool:
        """Check if a string ends with YYYYMMDD.

        :param index: Index string to check.
        :return: Whether it has YYYYMMDD suffix.
        """

        suffix = index[-8:]  # Last 8 chars should be YYYYMMDD
        try:
            pendulum.parse(suffix)
            return True
        except:
            return False

    @staticmethod
    def bucket_indices(indices: Iterable) -> dict:
        """Organises the indices into buckets, where each index in the same bucket has the same string before the
        YYYYMMDD suffix.

        :param indices: List of indices to bucket.
        :return: Dictionary of index buckets.
        """

        buckets = {}

        for index in indices:
            key = index[:-9]  # string before -YYYYMMDD
            buckets[key] = buckets.get(key, list())
            buckets[key].append(index)

        return buckets

    @staticmethod
    def sort_buckets(*, buckets: dict, order: KeepOrder) -> dict:
        """Sort the buckets into either increasing date or decreasing date sequences.

        :param buckets: Buckets to sort.
        :param order: Sorting order.
        :return: Sorted buckets.
        """

        sorted_buckets = {}

        for key, bucket in buckets.items():
            if order == KeepOrder.oldest:
                sorted_buckets[key] = sorted(bucket)
            else:
                sorted_buckets[key] = sorted(bucket, reverse=True)

        return sorted_buckets

    @staticmethod
    def get_deletion_list(*, buckets: dict, num_keep: int) -> List[str]:
        """Get a list of stale indices to delete.

        :param buckets: Buckets to get indices from.
        :param num_keep: Maximum number of indices to keep from each bucket.
        :return: List of indices to delete.
        """

        result = list()

        for bucket in buckets.values():
            result.extend(bucket[num_keep:])

        return result


class Elastic:
    def __init__(
        self,
        host: str = "http://elasticsearch:9200/",
        thread_count: int = cpu_count(),
        chunk_size: int = 10000,
        timeout: int = 30000,
        api_key_id: str = None,
        api_key: str = None,
    ):
        """Create an Elastic API client.

        :param host: the host including the hostname and port.
        :param thread_count: the number of threads to use when loading data into Elastic.
        :param chunk_size: the batch size for loading documents.
        :param timeout: the timeout in seconds.
        :param api_key_id: the Elastic API key id.
        :param api_key: the Elastic API key.
        """

        self.host = host
        self.thread_count = thread_count
        self.chunk_size = chunk_size
        self.timeout = timeout

        es_api_key = None
        if api_key_id is not None and api_key is not None:
            es_api_key = (api_key_id, api_key)

        self.es = Elasticsearch(hosts=[self.host], timeout=timeout, api_key=es_api_key)

    def query(self, index: str, query: Dict = None):
        if query is None:
            query = {"query": {"match_all": {}}}
        records = []
        for result in scan(self.es, query=query, index=index):
            records.append(result["_source"])

        return records

    def delete_index(self, index_id: str) -> None:
        """Delete an Elastic index.

        :param index_id: the index ID.
        :return: None.
        """

        try:
            self.es.indices.delete(index=index_id)
        except elasticsearch.NotFoundError as e:
            logging.warning(f"Index not found: {e}")

    def delete_indices(self, indices: List[str]):
        """Delete a list of Elastic indices.

        :param indices: List of Elastic indices to delete.
        """

        for index in indices:
            self.delete_index(index)

    def get_alias_indexes(self, alias_id: str) -> List:
        """Return the list of indexes associated with an alias.

        :param alias_id: the alias identifier.
        :return: the list of indexes.
        """

        index_ids = []
        try:
            indexes: Dict = self.es.indices.get(index=alias_id)
            for key, val in indexes.items():
                index_ids.append(key)
        except elasticsearch.exceptions.NotFoundError:
            pass

        return index_ids

    def index_documents(self, index_id: str, mappings: Dict, documents: Iterator[Dict]) -> bool:
        """Index documents supplied as an iterator.

        :param index_id: the index ID.
        :param mappings: the mappings supplied as a Dict.
        :param documents: the documents supplied as an iterator.
        :return: whether the documents were loaded successfully or not.
        """

        # Load mapping and create index
        body = {
            "settings": {
                "index": {"number_of_shards": 1, "number_of_replicas": 1},
                "analysis": {
                    "analyzer": {
                        "text_analyzer": {
                            "type": "custom",
                            "tokenizer": "standard",
                            "filter": ["lowercase", "asciifolding"],
                        }
                    }
                },
            },
            "mappings": mappings["mappings"],
        }
        self.es.indices.create(index=index_id, body=body, ignore=400)

        # Ingest data
        results = []
        for success, info in parallel_bulk(
            self.es,
            self._yield_bulk_documents(index_id, documents),
            refresh="wait_for",
            queue_size=self.thread_count,
            thread_count=self.thread_count,
            chunk_size=self.chunk_size,
            request_timeout=self.timeout,
        ):
            results.append(success)
            if not success:
                logging.error(f"Document failed to index: {success}")
            else:
                logging.debug(f"Indexed document: {success}")

        return all(results)

    def _yield_bulk_documents(self, index_id: str, documents: Iterator[Dict]):
        """Yields documents for indexing.

        :param index_id: the index ID.
        :param documents: the document iterator.
        :return: None.
        """

        for doc in documents:
            yield {"_op_type": "index", "_index": index_id, "_source": doc}

    def create_index(self, index: str, **kwargs):
        """Create a new Elastic index.

        :param index: Index to create.
        :param **kwargs: Any additional arguments to pass to the Elasticsearch client.
        """

        self.es.indices.create(index=index, **kwargs)

    def list_indices(self, index: str) -> List[str]:
        """List all indices matching the query name. The name can contain wildcards.
        :param index: Index name to query.
        :return: List of matching index names.
        """

        try:
            response = self.es.indices.get_alias(index=index)
            indices = response.keys()
            return list(indices)
        except:
            return list()

    def delete_stale_indices(self, *, index: str, keep_info: KeepInfo):
        """Keep a certain number of most recent releases (or oldest) for an index name.  The index name can
        have wildcard if you are not matching an exact index name. For example, test-name* will match all indices with
        the test-name prefix. The indices must end in the suffix -YYYYMMDD.  Indices not ending with -YYYYMMDD will be
        ignored.

        If more than one result is returned, the names will be clustered so that indices with the same name before the
        -YYYYMMDD suffix are together.  Each cluster will be sorted by the date in the suffix string, and the stale
        indices will be deleted.

        :param index: Index search string to delete stale versions of.
        :param keep_info: Metadata about how to order indices, and how many indices to keep.
        """

        # Get a list of indices matching our string.
        indices = self.list_indices(index)

        # Prune names that are too short.
        indices = list(filter(lambda x: len(x) > 9, indices))  # Length of -YYYYMMDD is 9

        # Prune names that dont have -YYYYMMDD suffix.
        indices = list(filter(StaleIndexCleaner.has_yyyymmdd_suffix, indices))

        # Put same prefixes into the same bucket. Use hash table.
        buckets = StaleIndexCleaner.bucket_indices(indices)

        # Sort each bucket. If reverse is false, sort descending, else sort ascending.
        sorted_buckets = StaleIndexCleaner.sort_buckets(buckets=buckets, order=keep_info.ordering)

        # Calculate indices to delete
        deletion_list = StaleIndexCleaner.get_deletion_list(buckets=sorted_buckets, num_keep=keep_info.num)

        # For each index in the tail, delete the index.
        self.delete_indices(deletion_list)
